import random
from typing import List, TypeVar, Optional
import pronouncing
from datamuse import datamuse
from .utils import *

api = datamuse.Datamuse()
str_or_list_of_str = TypeVar('str_or_list_of_str', str, List[str])


def rhymes(input_val: str_or_list_of_str, sample_size=None) -> List[str]:
    """Return a list of rhymes in randomized order for a given word if at least one can be found using the pronouncing
    module (which uses the CMU rhyming dictionary).

    :param input_val: the word or words in relation to which this function is looking up rhymes
    :param sample size: If provided, return a random sample of this many elements. If this number is greater than
                        the length of the rhyme list, then just return a shuffled copy of the rhyme list.
    """
    input_words = validate_str_or_list_of_str(input_val)
    rhyme_words: List[str] = []
    for input_word in input_words:
        rhyme_words.extend(filter_word_list([word for word in set(pronouncing.rhymes(input_word))]))
    return extract_sample(rhyme_words, sample_size=sample_size)


def rhyme(input_word: str) -> Optional[str]:
    """Return a random rhyme for a given word if at least one can be found using the pronouncing module (which uses
    the CMU rhyming dictionary).

    :param input_word: the word which this function is looking up a rhyme of
    """
    rhyme_list = rhymes(input_word)
    if len(rhyme_list):
        return next(iter(rhyme_list), None)
    return None


def extract_sample(word_list: list, sample_size: Optional[int] = None) -> list:
    """Returns a random sample from the word list or a shuffled copy of the word list.

    :param word_list: the list of words to extract the random sample of k
    :param sample_size: If this number is greater than the length of the word list, then just return a shuffled
                        copy of the word list.
    """
    if not sample_size or len(word_list) <= sample_size:
        return random.sample(word_list, k=len(word_list))
    else:
        sample: List[str] = []
        while len(sample) < sample_size and len(word_list) > 0:
            sample += [word for word in random.sample(word_list, k=sample_size) if word not in sample]
            word_list = [word for word in word_list if word not in sample]
        if sample_size < len(sample):
            return random.sample(sample, k=sample_size)
        return sample


def similar_sounding_words(input_val: str_or_list_of_str, sample_size: Optional[int] = 6,
                           datamuse_api_max: Optional[int] = 50) -> list:
    """Return a list of similar sounding words to a given word, in randomized order, if at least one can be found using
    Datamuse API.

    :param input_val: the word or words in relation to which this function is looking up similar sounding words
    :param sample_size: If provided, return a random sample of this many elements. If this number is greater than the
                        length of the API results, then just return a shuffled copy of the filtered API results.
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's results
                             are always sorted from most to least similar sounding (according to a numeric score
                             provided by Datamuse), hence by using both parameters, one can control the size of both
                             the sample pool and the sample size.
    """
    input_words = validate_str_or_list_of_str(input_val)
    ss_words: List[str] = []
    for input_word in input_words:
        response = api.words(sl=input_word, max=datamuse_api_max) if datamuse_api_max else api.words(sl=input_word)
        exclude_words = input_words + ss_words
        ss_words.extend(filter_word_list([obj['word'] for obj in response], exclude_words=exclude_words))
    return extract_sample(ss_words, sample_size=sample_size)


def similar_sounding_word(input_word: str, datamuse_api_max: Optional[int] = 20) -> Optional[str]:
    """Return a random similar sounding word for a given word if at least one can be found using the Datamuse API.

    :param input_word: the word which this function is looking up a similar sounding words of
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's results are
                             always sorted from most to least similar sounding (according to a numeric score provided
                             by Datamuse).
    """
    return next(iter(similar_sounding_words(input_word, sample_size=1, datamuse_api_max=datamuse_api_max)), None)


def similar_meaning_words(input_val: str_or_list_of_str, sample_size: Optional[int] = 6,
                          datamuse_api_max: Optional[int] = 20) -> list:
    """Return a list of similar meaning words to a given word, in randomized order, if at least one can be found using
    Datamuse API.

    :param input_val: the word or words in relation to which this function is looking up similar meaning words
    :param sample_size: If provided, return a random sample of this many elements. If this number is greater than the
                        length of the API results, then just return a shuffled copy of the filtered API results.
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's results
                             are always sorted from most to least similar meaning (according to a numeric score
                             provided by Datamuse), hence by using both parameters, one can control the size of both
                             the sample pool and the sample size.
    """
    input_words = validate_str_or_list_of_str(input_val)
    sm_words: List[str] = []
    for input_word in input_words:
        response = api.words(ml=input_word, max=datamuse_api_max) if datamuse_api_max else api.words(ml=input_word)
        exclude_words = sm_words.copy()
        sm_words.extend(filter_word_list([obj['word'] for obj in response], spellcheck=False,
                                         exclude_words=exclude_words))
    return extract_sample(sm_words, sample_size=sample_size)


def similar_meaning_word(input_word: str, datamuse_api_max: Optional[int] = 10) -> Optional[str]:
    """Return a random similar meaning word for a given word if at least one can be found using the Datamuse API.

    :param input_word: the word which this function is looking up a similar meaning words of
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's results are
                             always sorted from most to least similar meaning (according to a numeric score provided
                             by Datamuse).
    """
    return next(iter(similar_meaning_words(input_word, sample_size=1, datamuse_api_max=datamuse_api_max)), None)


def contextually_linked_words(input_val: str_or_list_of_str, sample_size: Optional[int] = 6,
                              datamuse_api_max: Optional[int] = 20) -> list:
    """Return a list of words that frequently appear within the same document as a given word, in randomized order,
    if at least one can be found using the Datamuse API.

    :param input_val: the word or words in relation to which this function is looking up contextually linked words
    :param sample_size: If provided, return a random sample of this many elements. If this number is greater than the
                        length of the API results, then just return a shuffled copy of the filtered API results.
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's results
                             are always sorted from most to least frequently coappearing (according to a numeric score
                             provided by Datamuse), hence by using both parameters, one can control the size of both
                             the sample pool and the sample size.
    """
    input_words = validate_str_or_list_of_str(input_val)
    cl_words: List[str] = []
    for input_word in input_words:
        validate_word(input_word)
        response = api.words(rel_trg=input_word, max=datamuse_api_max) if datamuse_api_max else \
            api.words(rel_trg=input_word)
        exclude_words = cl_words.copy()
        # Spellcheck removes proper nouns so don't.
        cl_words.extend(filter_word_list([obj['word'] for obj in response], spellcheck=False,
                                         exclude_words=exclude_words))
    return extract_sample(cl_words, sample_size=sample_size)


def contextually_linked_word(input_word: str, datamuse_api_max: Optional[int] = 10) -> Optional[str]:
    """Return a random word that frequently appear within the same document as a given word if at least one can be found
    using the Datamuse API.

    :param input_word: the word which this function is looking up a contextually linked words to
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's results are
                             always sorted from most to least similar sounding (according to a numeric score provided
                             by Datamuse).
    """
    return next(iter(contextually_linked_words(input_word, sample_size=1, datamuse_api_max=datamuse_api_max)), None)


def frequently_following_words(input_val: str_or_list_of_str, sample_size: Optional[int] = 8,
                               datamuse_api_max: Optional[int] = None) -> list:
    """Return a list of words that frequently follow the given word, in randomized order, if at least one can be found
    using the Datamuse API.

    :param input_val: the word or words in relation to which this function is looking up frequently following words
    :param sample_size: If provided, return a random sample of this many elements. If this number is greater than
                             the length of the API results, then just return a shuffled copy of the filtered API
                             results.
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's
                                  results are always sorted from most to least frequently coappearing (according to a
                                  numeric score provided by Datamuse), hence by using both parameters, one can control
                                  the size of both the sample pool and the sample size.
    """
    input_words = validate_str_or_list_of_str(input_val)
    ff_words: List[str] = []
    for input_word in input_words:
        response = api.words(lc=input_word, max=datamuse_api_max) if datamuse_api_max else api.words(lc=input_word)
        # Filter but don't use spellcheck -- it removes important words (for the markov chain use case) like 'of'
        exclude_words = ff_words.copy()
        ff_words.extend(filter_word_list([obj['word'] for obj in response], spellcheck=False,
                                         exclude_words=exclude_words))
        random.shuffle(ff_words)
    if sample_size and sample_size > 4:
        # Pick 3 at random from the top X rarest and the rest from the whole
        # Slice one list of api results using the default order and another using a rarity baeed order
        if not datamuse_api_max:
            ending_index = 20
        elif datamuse_api_max % 2 == 1:
            ending_index = datamuse_api_max + 1
        else:
            ending_index = datamuse_api_max
        return extract_sample(ff_words[:ending_index], sample_size=sample_size - 3) + \
            extract_sample(sort_by_rarity(ff_words)[:ending_index], sample_size=3)
    return extract_sample(ff_words, sample_size=sample_size)  # Standard sampling


def frequently_following_word(input_word, datamuse_api_max=10) -> Optional[str]:
    """Return a random word that frequently follows the given word if at least one can be found using the Datamuse API.

    :param input_word: the word which this function is looking up a frequently following word of
    :param datamuse_api_max: specifies the maximum number of results returned by the API. The API client's results are
                             always sorted from most to least similar sounding (according to a numeric score provided
                             by Datamuse).
    """
    result: Optional[str] = next(iter(frequently_following_words(input_word, sample_size=1,
                                                                 datamuse_api_max=datamuse_api_max)), None)
    return result


def phonetically_related_words(input_val: str_or_list_of_str, sample_size=None, datamuse_api_max=50,
                               limit_results_per_input_word: Optional[int] = None) -> list:
    """Returns a list of rhymes and similar sounding words to a word or list of words.

    :param input_val: the word or words in relation to which this function is looking up phonetically related words
    :param sample_size: If provided, pass this argument to the functions rhymes and similar_sounding_words so that
                         twice this number of elements are returned by this function. If not provided, the function
                         will return all rhymes plus however many API results similar_sounding_words.
    :param datamuse_api_max: specifies how many API results can be returned by the API client when fetching similar
                        meaning words.
    """
    input_words = validate_str_or_list_of_str(input_val)
    results: List[str] = []
    for word in input_words:
        results.extend(rhymes(word, sample_size=limit_results_per_input_word))
        exclude_words = results.copy()
        nonrhymes = filter_word_list(similar_sounding_words(
            word, sample_size=sample_size, datamuse_api_max=datamuse_api_max), exclude_words=exclude_words)
        results.extend(nonrhymes[:limit_results_per_input_word])
    results = extract_sample(results, sample_size=sample_size)
    return results


def related_rare_words(input_val: str_or_list_of_str, sample_size: Optional[int] = 8,
                       rare_word_population_max: int = 20) -> list:
    """Return a random sample of rare related words to a given word. The words can be related phonetically,
    contextually, or by meaning).

    :param input_val: the word or words in relation to which this function is looking up related rare words
    :param sample_size: If provided, return a random sample of this many elements. If this number is greater than
                        the length of rare word population size, then just return a shuffled copy of that.
    :param rare_word_population_max: specifies the maximum number of related words to subsample from per word.
    `                                The rare word population is sorted from rarest to most common. If sample_size is
                                     null, the max results returned by this function is 2 times this number.
    """
    input_words = validate_str_or_list_of_str(input_val)
    results: List[str] = []
    for input_word in input_words:
        related_words = phonetically_related_words(input_word)
        related_words.extend(word for word in contextually_linked_words(
            input_word, sample_size=None, datamuse_api_max=100) if word not in related_words)
        related_words.extend(word for word in similar_meaning_words(
            input_word, sample_size=None, datamuse_api_max=100) if word not in related_words)
        related_words = [word for word in related_words if not too_similar(input_word, word)]
        results.extend(sort_by_rarity(related_words)[:rare_word_population_max])
    return extract_sample(results, sample_size=sample_size)


def related_rare_word(input_word: str, rare_word_population_max: int = 10) -> Optional[str]:
    """Return a random rare related word to a given word. The word can be related phonetically, contextually, or by
    meaning).

    :param input_word: the word which this function is looking up related rare words to
    :param rare_word_population_max: specifies the maximum number of related words to subsample from. The rare word
                                    population is sorted from rarest to most common.
    """
    return next(iter(related_rare_words(input_word, sample_size=1,
                                        rare_word_population_max=rare_word_population_max)), None)


import pkgutil
import platform
import random
import re
import hunspell
from consolemenu.screen import Screen
from typing import List, TypeVar
from wordfreq import word_frequency


def setup_spellchecker():
    if platform.system() == 'Windows':
        raise Exception('Your OS is not currently supported.')
    elif platform.system() == 'Darwin':
        try:
            return hunspell.HunSpell('/Library/Spelling/en_US.dic', '/Library/Spelling/en_US.aff')
        except Exception:
            raise Exception('This module requires the installation of the hunspell dictionary.')
    else:
        try:
            return hunspell.HunSpell('/usr/share/hunspell/en_US.dic', '/usr/share/hunspell/en_US.aff')
        except Exception:
            raise Exception('This module requires the installation of the hunspell dictionary.')


hobj = setup_spellchecker()
str_or_list_of_str = TypeVar('str_or_list_of_str', str, List[str])

def get_input_words():
    prompt = 'To generate a poem, type some words separated by commas or spaces, and then press enter.\n\n'

    input_words = []
    while len(input_words) == 0:
        inp = Screen().input(prompt=prompt)
        input_words = [word for word in re.split(r'[\s,]', inp) if word]
    return input_words


def get_random_color(threshold=.85):
    """Gets a random color -- 2/3 of the rgb values must be below the threshold value"""
    r, g, b = 1, 1, 1
    while (1 - r <= threshold and 1 - g <= threshold) or (1 - g <= threshold and 1 - b <= threshold) or \
            (1 - r <= threshold and 1 - b <= threshold):
        r, g, b = random.random(), random.random(), random.random()
    return r, g, b


def validate_str(input_val, msg='Not a string'):
    """Validate the input argument by checking if it is a string.

    :param input_val: the value to validate
    :param msg: the message to display if a ValueError is thrown
    """
    if not isinstance(input_val, str):
        raise ValueError(msg)


def validate_str_list(input_val, msg='Not a list'):
    """Validate the input parameter by checking if it is a list of strings.

    :param input_val: the value to validate
    :param msg: the message to display if a ValueError is thrown
    """
    if not isinstance(input_val, list):
        raise ValueError(msg)
    for i, elem in enumerate(input_val):
        if not isinstance(elem, str):
            raise ValueError(f'Element {i + 1} not a string')


def validate_str_or_list_of_str(input_val) -> List[str]:
    """Validate the input parameter by checking if it is a string or a list of strings.

    :param input_val: the value to validate
    """
    if isinstance(input_val, str):
        return [input_val]
    elif isinstance(input_val, list):
        validate_str_list(input_val, msg='Must provide a string or list of strings')
        return input_val
    else:
        raise ValueError('Must provide a string or list of strings')


def has_invalid_characters(string):
    """Check if the string has unpermitted characters: whitespace, digits, and hyphens.

    :param string: the string to check for invalid characters in
    """
    return bool(re.search(r"[\s\d\-\']", string))


def validate_word(input_val):
    """Check whether the input argument is a word.

    :param input_val: the word to validate
    """
    validate_str(input_val)
    if has_invalid_characters(input_val):
        raise ValueError('Word may not contain digits, spaces, or special characters.')


def filter_word(string, spellcheck=True, exclude_words=[], word_frequency_threshold=4e-08):
    """Filter out a word if it is too short, has invalid characters, is too archaic, or (optionally) cannot be found in
    a spelling dictionary.

    :param string: the string to check against
    :param spellcheck: Use a spelling dictionary as filter. This helps eliminate abbreviations, proper nouns, and
                       Internet slang--sometimes this is not desirable. It also eliminates many short stopwords like
                       'of' for some reason.
    :param exclude_words: list of words to filter out
    :param word_frequency_threshold: how frequently the word appears in the word_frequency package's corpus -- filter
                                     out word if less frequent than this threshold
    """
    # Datamuse is built from webscraping and occasionally returns offensive and oppressive language, which I am here
    # adding to filter out. Although there is an appropriate and even critical way for humans to write poetry using some
    # of these words that might be considered edge cases (e.g. Hottentot), a stochastic text generator does not have
    # a historical sense to do that, so I have decided to exclude these.
    unfitting_words = pkgutil.get_data('generativepoetry', 'data/hate_words.txt').decode("utf-8").splitlines()
    unfitting_words.extend(pkgutil.get_data('generativepoetry', 'data/abbreviations_etc.txt').decode("utf-8")
                           .splitlines())
    exclude_words.extend(unfitting_words)  # Some words Datamuse tends to return that disruptive poetic flow
    validate_str(string)
    if len(string) < 3:
        return False
    if has_invalid_characters(string):
        return False
    if word_frequency(string, 'en') < word_frequency_threshold:
        return False
    if spellcheck and not hobj.spell(string):
        return False
    if string in exclude_words:
        return False
    return True


def filter_word_list(word_list: List[str], spellcheck: bool = True, exclude_words: List[str] = []) -> List[str]:
    """Filter a list of words using the filter_word method.

    :param word_list: list of words to filter
    :param spellcheck (bool) -- Use a spelling dictionary as filter (helps eliminate abbreviations and Internet slang).
    """
    results: List[str] = list(
        filter(
            lambda word: filter_word(word, spellcheck=spellcheck, exclude_words=exclude_words), word_list
        )
    )
    return results


def sort_by_rarity(word_list: List[str]) -> List[str]:
    if len(word_list) <= 1:
        return word_list
    return sort_by_rarity(
        [word for word in word_list[1:] if word_frequency(word, 'en') < word_frequency(word_list[0], 'en')]
    ) + [word_list[0]] + \
           sort_by_rarity(
               [word for word in word_list[1:] if word_frequency(word, 'en') >= word_frequency(word_list[0], 'en')])


def too_similar(word1: str, comparison_val: str_or_list_of_str) -> bool:
    """Check whether or not two words are too similar to follow one another in a poem, e.g. if one is the other plus s.
    :param word1: the first word to compare
    :param comparison_val: a word or list of words to compare against
    """
    validate_str(word1)
    comparison_words = validate_str_or_list_of_str(comparison_val)
    for word2 in comparison_words:
        if len(word1) == 0 or len(word2) == 0:
            return False
        if word1 == word2:
            return True
        if word1 + 's' == word2 or word2 + 's' == word1:  # Plural, probably
            return True
        if word1 + 'ly' == word2 or word2 + 'ly' == word1:  # Adverb form of an adjective
            return True
        # Perhaps the latter two checks could still be done efficiently with lemmatization
        if (len(word1) > 2 and len(word2) > 2) and ((word1[-2] == 'e' and word2 + 'd' == word1) or
            (word2[-2] == 'e' and word1 + 'd' == word2)):  # Past tense
            return True
        if ((len(word1) > 5 and len(word2) > 2) or (len(word2) > 5 and len(word1) > 2)) and \
                ((word1[-3:] == 'ing' and word2 + 'ing' == word1) or (word2[-3:] == 'ing' and word1 + 'ing' == word2)):
            # Gerunds
            return True
        too_similar_case = ['the', 'thee', 'them']
        if word1 in too_similar_case and word2 in too_similar_case:
            return True
    return False


def correct_a_vs_an(phrase_as_list: List[str]) -> List[str]:
    consonants = 'bcdfghjklmnpqrstvwxyz'
    vowels = 'aeoiu'
    last_word = None
    for i, word in enumerate(phrase_as_list):
        if last_word == 'a':
            if word[0] in vowels:
                phrase_as_list[i - 1] = 'an'
        elif last_word == 'an':
            if word[0] in consonants:
                phrase_as_list[i - 1] = 'a'
        last_word = word
    return phrase_as_list



import itertools
import os
import re
import unittest
from unittest.mock import patch
from generativepoetry.lexigen import *
from generativepoetry.pdf import *
from generativepoetry.poemgen import *
from generativepoetry.utils import *


class TestUtils(unittest.TestCase):

    def test_setup_spellchecker(self):
        self.assertIsNotNone(setup_spellchecker())

    def test_get_random_color(self):
        for i in range(4):
            rgb = get_random_color()
            red_over = rgb[0] >= .85
            green_over = rgb[1] >= .85
            blue_over = rgb[2] >= .85
            self.assertFalse(red_over and green_over) or (red_over and blue_over) or (green_over and blue_over)
        for i in range(4):
            threshold = .25
            rgb = get_random_color()
            red_over = rgb[0] >= threshold
            green_over = rgb[1] >= threshold
            blue_over = rgb[2] >= threshold
            self.assertFalse(red_over and green_over) or (red_over and blue_over) or (green_over and blue_over)

    def test_get_input_words(self):
        test_input = 'trout mask replica'
        with patch('builtins.input', return_value=test_input):
            self.assertEqual(get_input_words(), ['trout', 'mask', 'replica'])
        test_input = 'fish,paperclip,atom,benign'
        with patch('builtins.input', return_value=test_input):
            self.assertEqual(get_input_words(), ['fish', 'paperclip', 'atom', 'benign'])
        test_input = 'fish, paperclip, atom, benign'
        with patch('builtins.input', return_value=test_input):
            self.assertEqual(get_input_words(), ['fish', 'paperclip', 'atom', 'benign'])

    def test_validate_str(self):
        self.assertRaises(ValueError, lambda: validate_str(2))
        self.assertRaises(ValueError, lambda: validate_str(2.5))
        self.assertRaises(ValueError, lambda: validate_str(False))
        self.assertRaises(ValueError, lambda: validate_str(None))
        validate_str('lingo')

    def test_validate_str_list(self):
        self.assertRaises(ValueError, lambda: validate_str_list(2))
        self.assertRaises(ValueError, lambda: validate_str_list(2.5))
        self.assertRaises(ValueError, lambda: validate_str_list(False))
        self.assertRaises(ValueError, lambda: validate_str_list(None))
        self.assertRaises(ValueError, lambda: validate_str_list('not a list'))
        self.assertRaises(ValueError, lambda: validate_str_list(['a', 'b', None]))
        validate_str_list(['a', 'b', 'c'])

    def test_validate_str_or_list_of_str(self):
        self.assertRaises(ValueError, lambda: validate_str_list(2))
        self.assertRaises(ValueError, lambda: validate_str_list(2.5))
        self.assertRaises(ValueError, lambda: validate_str_list(False))
        self.assertRaises(ValueError, lambda: validate_str_list(None))
        self.assertRaises(ValueError, lambda: validate_str_list(['a', 'b', None]))
        validate_str('deciduous')
        validate_str_list(['anodyne', 'bolo', 'cdrkssdjak'])

    def test_has_invalid_characters(self):
        self.assertTrue(has_invalid_characters('gh0st'))
        self.assertTrue(has_invalid_characters('compound word'))
        self.assertTrue(has_invalid_characters('compound-word'))
        self.assertTrue(has_invalid_characters("apostrophe'"))
        self.assertFalse(has_invalid_characters('espousal'))

    def test_too_similar(self):
        self.assertRaises(ValueError, lambda: too_similar(None, 25.2))
        self.assertRaises(ValueError, lambda: too_similar('string', 25))
        self.assertRaises(ValueError, lambda: too_similar(list(), 'beans'))
        self.assertFalse(too_similar('self', 'other'))
        self.assertTrue(too_similar('dog', 'dog'))
        self.assertTrue(too_similar('dog', 'dogs'))
        self.assertTrue(too_similar('dogs', 'dog'))
        self.assertTrue(too_similar('spherical', 'spherically'))
        self.assertTrue(too_similar('spherically', 'spherical'))
        self.assertTrue(too_similar('riposte', 'riposted'))
        self.assertTrue(too_similar('riposted', 'riposte'))
        self.assertTrue(too_similar('riposte', ['dogs', 'mushroom', 'riposted']))
        self.assertFalse(too_similar('riposte', ['dogs', 'mushroom', 'quails']))
        self.assertTrue(too_similar('thee', 'the'))
        self.assertTrue(too_similar('thee', 'the'))

    def test_filter_word(self):
        self.assertFalse(filter_word('an'))
        self.assertFalse(filter_word('nonexistentword'))
        self.assertFalse(filter_word('errantry'))  # 1.51e-08 so below threshold
        self.assertTrue(filter_word('crepuscular'))  # 7.41e-08 so OK
        self.assertTrue(filter_word('puppy'))
        self.assertFalse(filter_word('thew'))  # from the annoying words list

    def test_filter_word_list(self):
        word_list = ['the', 'crepuscular', 'dogs']
        self.assertEqual(filter_word_list(word_list), word_list) ## All spelled correctly
        word_list = ['the', 'underworld', 'gh0st', 'errantry', 'an']
        valid_words = ['the', 'underworld']
        self.assertEqual(filter_word_list(word_list), valid_words)
        word_list = ['araignment', 'arraignment', 'dynosaur', 'dinosaur']
        correctly_spelled_word_list = ['arraignment', 'dinosaur']
        self.assertEqual(filter_word_list(word_list), correctly_spelled_word_list)
        exclude_words = ['diamond', 'dinosaur']
        self.assertEqual(filter_word_list(word_list, exclude_words=exclude_words), ['arraignment'])

    def test_sort_by_rarity(self):
        unsorted_words = ['cat', 'catabasis', 'hue', 'corncob',  'the', 'Catalan', 'errant']
        correctly_sorted_words = ['catabasis', 'corncob', 'errant', 'hue', 'Catalan', 'cat', 'the']
        self.assertEqual(sort_by_rarity(unsorted_words), correctly_sorted_words)

    def correct_a_vs_a_vs_an(self):
        needs_no_correction = ['an', 'obscure' 'elephant' 'and' 'a' 'wandering' 'heliotrope', 'see', 'a', '3']
        self.assertEqual(correct_a_vs_an(needs_no_correction), needs_no_correction)
        needs_correction = ['a', 'obscure' 'elephant' 'and' 'an' 'wandering' 'heliotrope', 'see', 'an', '3']
        self.assertEqual(correct_a_vs_an(needs_no_correction), needs_no_correction)


class TestLexigen(unittest.TestCase):
    def test_rhymes(self):
        self.assertEqual(rhymes('metamorphosis'), [])
        rhymes_with_clouds = ['crowds', 'shrouds']
        results = rhymes('clouds')
        self.assertEqual(sorted(results), rhymes_with_clouds)
        rhymes_with_sprouting = ['doubting', 'flouting', 'grouting', 'outing', 'pouting', 'rerouting', 'routing',
                                 'scouting', 'shouting', 'spouting', 'touting']
        self.assertEqual(sorted(rhymes('sprouting', sample_size=None)), rhymes_with_sprouting)
        results = rhymes('sprouting')
        self.assertNotIn('sprouting', results)
        self.assertEqual(sorted(results), rhymes_with_sprouting)
        results = rhymes('sprouting', sample_size=6)
        self.assertEqual(len(results), 6)
        self.assertTrue(set(rhymes_with_sprouting).issuperset(set(results)))
        rhymes_with_either = sorted(rhymes_with_clouds + rhymes_with_sprouting)
        self.assertEqual(sorted(rhymes(['sprouting', 'clouds'], sample_size=None)), rhymes_with_either)
        results = rhymes(['clouds', 'sprouting'], sample_size=6)
        self.assertEqual(len(results), 6)
        self.assertTrue(set(rhymes_with_either).issuperset(set(results)))

    def test_rhyme(self):
        self.assertIsNone(rhyme('metamorphosis'))
        self.assertIn(rhyme('sprouting'), rhymes('sprouting'))

    def test_extract_sample(self):
        self.assertEqual(extract_sample([], sample_size=100), [])
        self.assertEqual(extract_sample(['a'], sample_size=100), ['a'])
        self.assertEqual(sorted(extract_sample(['a','b','c'], sample_size=3)), ['a','b','c'])
        sample = extract_sample(['a','b','c','d','e','f'], sample_size=4)
        self.assertNotEqual(sorted(sample), ['a','b','c','d','e','f'])
        self.assertTrue(set(['a','b','c','d','e','f']).issuperset(set(sample)))

    def test_similar_sounding_words(self):
        similar_sounding_to_homonym_words = ['hastening', 'heightening', 'hominid', 'hominy', 'homonyms', 'summoning',
                                             'synonym']
        self.assertEqual(sorted(similar_sounding_words('homonym', sample_size=None)), similar_sounding_to_homonym_words)
        results = similar_sounding_words('homonym')
        self.assertEqual(len(results), 6)
        self.assertTrue(set(similar_sounding_to_homonym_words).issuperset(set(results)))
        similar_sounding_to_ennui_words = ['anew', 'any', 'emcee', 'empty']
        all_similar_sounding_words = sorted(similar_sounding_to_homonym_words + similar_sounding_to_ennui_words)
        self.assertEqual(sorted(similar_sounding_words(['homonym', 'ennui'], sample_size=None)),
                         all_similar_sounding_words)
        results = similar_sounding_words(['homonym', 'ennui'])
        self.assertEqual(len(results), 6)
        self.assertTrue(set(all_similar_sounding_words).issuperset(set(results)))

    def test_similar_sounding_word(self):
        self.assertIsNone(similar_sounding_word('voodoo'))
        all_similar_sounding_words = ['hastening', 'heightening', 'hominid', 'hominy', 'homonym', 'homonyms',
                                      'summoning', 'synonym']  # Using this to save API call in test
        self.assertIn(similar_sounding_word('homonym'), all_similar_sounding_words)

    def test_similar_meaning_words(self):
        self.assertEqual(similar_meaning_words('nonexistentword'), [])
        similar_meaning_to_vampire_words = ['bats', 'bloodsucker', 'clan', 'demon', 'ghoul', 'james', 'kind', 'lamia',
                                            'lycanthrope', 'lycanthropy', 'shane', 'shapeshifter', 'succubus', 'undead',
                                            'vamp', 'vampirism', 'werewolf', 'witch', 'wolfman', 'zombie']
        results = similar_meaning_words('vampire', sample_size=None)
        self.assertEqual(len(results), 20)
        self.assertEqual(sorted(results), similar_meaning_to_vampire_words)
        results = similar_meaning_words('vampire', sample_size=None, datamuse_api_max=None)
        self.assertGreater(len(results), 20)
        self.assertTrue(set(results).issuperset(set(similar_meaning_to_vampire_words)))
        results = similar_meaning_words('vampire', sample_size=6)
        self.assertEqual(len(results), 6)
        self.assertTrue(set(similar_meaning_to_vampire_words).issuperset(set(results)))
        similar_meaning_to_gothic = ['goth', 'hard', 'eldritch', 'unusual', 'spooky', 'rococo', 'minimalist', 'folky',
                                     'lovecraftian', 'strange', 'baroque', 'creepy', 'medieval', 'mediaeval']
        similar_meaning_to_either = sorted(similar_meaning_to_vampire_words + similar_meaning_to_gothic)
        self.assertEqual(sorted(similar_meaning_words(['vampire', 'gothic'], sample_size=None)),
                         similar_meaning_to_either)
        results = similar_meaning_words(['vampire', 'gothic'])
        self.assertEqual(len(results), 6)
        self.assertTrue(set(similar_meaning_to_either).issuperset(set(results)))

    def test_similar_meaning_word(self):
        self.assertIsNone(similar_meaning_word('nonexistentword'))
        similar_meaning_to_vampire_words = ['bats', 'bloodsucker', 'clan', 'demon', 'ghoul', 'james', 'kind', 'lamia',
                                            'lycanthrope', 'lycanthropy', 'shane', 'shapeshifter', 'succubus', 'undead',
                                            'vamp', 'vampirism', 'werewolf', 'witch', 'wolfman', 'zombie']
        self.assertIn(similar_meaning_word('vampire'), similar_meaning_to_vampire_words)

    def test_contextually_linked_words(self):
        self.assertEqual(contextually_linked_words('nonexistentword'), [])
        contextually_linked_to_metamorphosis = ['budding', 'cocoon', 'duff', 'frogs', 'gills', 'hatching', 'juvenile',
                                                'kafka', 'lamprey', 'larva', 'metamorphose', 'narcissus', 'nymph',
                                                'polyp', 'polyps', 'pupa', 'pupal', 'salamander', 'starfish', 'tadpole']
        results = contextually_linked_words('metamorphosis', sample_size=None)
        self.assertEqual(len(results), 20)
        self.assertEqual(sorted(results), contextually_linked_to_metamorphosis)
        results = contextually_linked_words('metamorphosis', sample_size=None, datamuse_api_max=None)
        self.assertGreater(len(results), 20)
        self.assertTrue(set(results).issuperset(set(contextually_linked_to_metamorphosis)))
        results = contextually_linked_words('metamorphosis', sample_size=6)
        self.assertEqual(len(results), 6)
        self.assertTrue(set(contextually_linked_to_metamorphosis).issuperset(set(results)))
        contextually_linked_to_crepuscular = ['foraging', 'dusk', 'habits', 'twilight', 'diurnal', 'rays', 'dens',
                                              'forage', 'insects', 'nocturnal', 'overcast', 'predation', 'skipper',
                                              'sunset', 'moths', 'dawn', 'rodents', 'daylight', 'mating']
        contextually_linked_to_either = sorted(contextually_linked_to_crepuscular +
                                               contextually_linked_to_metamorphosis)
        self.assertEqual(sorted(contextually_linked_words(['crepuscular', 'metamorphosis'], sample_size=None)),
                         contextually_linked_to_either)
        results = contextually_linked_words(['crepuscular', 'metamorphosis'])
        self.assertEqual(len(results), 6)
        self.assertTrue(set(contextually_linked_to_either).issuperset(set(results)))

    def test_contextually_linked_word(self):
        self.assertIsNone(contextually_linked_word('nonexistentword'))
        contextually_linked_to_metamorphosis = ['kafka', 'lamprey', 'larva', 'metamorphose', 'narcissus', 'polyp',
                                                'polyps', 'pupa', 'pupal', 'tadpole']
        self.assertIn(contextually_linked_word('metamorphosis'), contextually_linked_to_metamorphosis)

    def test_frequently_following_words(self):
        self.assertEqual(frequently_following_words('nonexistentword'), [])
        frequently_following_magic = ['about', 'against', 'among', 'and', 'angle', 'are', 'art', 'arts', 'box',
                                      'bullet', 'bullets', 'but', 'can', 'carpet', 'charm', 'charms', 'circle', 'city',
                                      'could', 'flute', 'for', 'formula', 'formulas', 'from', 'had', 'has', 'influence',
                                      'into', 'johnson', 'key', 'kingdom', 'lamp', 'lantern', 'marker', 'markers',
                                      'may', 'mirror', 'moment', 'mountain', 'name', 'number', 'numbers', 'potion',
                                      'power', 'powers', 'realism', 'ring', 'rites', 'school', 'show',  'spell',
                                      'spells', 'square', 'squares', 'sword', 'than', 'that', 'the', 'touch', 'trick',
                                      'tricks', 'wand', 'was', 'were', 'when', 'which', 'will', 'with', 'word', 'words',
                                      'world', 'would']
        results = frequently_following_words('magic', sample_size=None)
        self.assertEqual(sorted(results), frequently_following_magic)
        results = frequently_following_words('magic', sample_size=None, datamuse_api_max=None)
        self.assertGreater(len(results), 20)
        self.assertTrue(set(results).issuperset(set(frequently_following_magic)))
        results = frequently_following_words('magic', sample_size=6)
        self.assertEqual(len(results), 6)
        self.assertTrue(set(frequently_following_magic).issuperset(set(results)))
        frequently_following_Dadaist = ['activities', 'and', 'art', 'artist', 'artists', 'collage', 'collages',
                                        'experiments', 'group', 'hugo', 'influence', 'kurt', 'manifesto', 'marcel',
                                        'movement', 'nihilism', 'painter', 'performance', 'play', 'poem', 'poems',
                                        'poet', 'poetry', 'poets', 'raoul', 'review', 'spirit', 'tradition', 'tristan',
                                        'who', 'work', 'works']
        ff_either = sorted(frequently_following_magic + [word for word in frequently_following_Dadaist
                                                         if word not in frequently_following_magic])
        self.assertEqual(sorted(frequently_following_words(['magic', 'Dadaist'], sample_size=None)), ff_either)
        results = frequently_following_words(['magic', 'Dadaist'])
        self.assertEqual(len(results), 8)
        self.assertTrue(set(ff_either).issuperset(set(results)))

    def test_frequently_following_word(self):
        self.assertIsNone(contextually_linked_word('nonexistentword'))
        frequently_following_Dadaist = ['activities', 'and', 'art', 'artist', 'artists', 'collage', 'collages',
                                        'experiments', 'group', 'hugo', 'influence', 'kurt', 'manifesto', 'marcel',
                                        'movement', 'nihilism', 'painter', 'performance', 'play', 'poem', 'poems',
                                        'poet', 'poetry', 'poets', 'raoul', 'review', 'spirit', 'tradition', 'tristan',
                                        'who', 'work', 'works']
        self.assertIn(frequently_following_word('Dadaist'), frequently_following_Dadaist)

    def test_phonetically_related_words(self):
        self.assertRaises(ValueError, lambda: phonetically_related_words(2))
        self.assertRaises(ValueError, lambda: phonetically_related_words(2.5))
        self.assertRaises(ValueError, lambda: phonetically_related_words(False))
        self.assertRaises(ValueError, lambda: phonetically_related_words(None))
        self.assertRaises(ValueError, lambda: phonetically_related_words(['a', 'b', None]))
        pr_to_poet = ['inchoate', 'opiate', 'payout', 'pet', 'peyote', 'pit', 'poached', 'poets', 'poked', 'post',
                      'putt']
        self.assertEqual(sorted(phonetically_related_words('poet', sample_size=None)), pr_to_poet)
        results = phonetically_related_words('poet', sample_size=5)
        self.assertEqual(len(sorted(results)), 5)
        self.assertTrue(set(sorted(pr_to_poet)).issuperset(set(results)))
        expected_pr_words = sorted(pr_to_poet + ['eon', 'gnawing', 'knowing', 'kneeing', 'naan', 'non', 'noun'])
        self.assertEqual(sorted(phonetically_related_words(['poet', 'neon'], sample_size=None)), expected_pr_words)

    def test_related_rare_words(self):
        self.assertRaises(ValueError, lambda: related_rare_words(2))
        self.assertRaises(ValueError, lambda: related_rare_words(2.5))
        self.assertRaises(ValueError, lambda: related_rare_words(False))
        self.assertRaises(ValueError, lambda: related_rare_words(None))
        rr_to_comical = ['absurdist', 'antic', 'artless', 'campy', 'canticle', 'cliched', 'clownish', 'cockle',
                         'cringeworthy', 'hackneyed', 'histrionic', 'humourous', 'jokey', 'parodic', 'puerile',
                         'risible', 'sophomoric', 'surrealistic', 'uneconomical', 'uproarious']
        results = related_rare_words('comical', sample_size=None)
        self.assertEqual(len(results), 20)
        self.assertEqual(sorted(results), rr_to_comical)
        results = related_rare_words('comical', sample_size=None, rare_word_population_max=None)
        self.assertGreater(len(results), 20)
        self.assertTrue(set(results).issuperset(set(rr_to_comical)))
        results = related_rare_words('comical', sample_size=6)
        self.assertEqual(len(results), 6)
        self.assertTrue(set(results).issuperset(set(results)))

        rr_to_dinosaur = ['allosaurus', 'apatosaurus', 'archaeopteryx', 'brachiosaurus', 'clade', 'crocodilian',
                          'diplodocus', 'dodos', 'humidor', 'ichthyosaur', 'iguanodon', 'megafauna', 'peccary',
                          'pterosaur', 'robustus', 'sauropod', 'stevedore', 'theropod', 'trilobite', 'tyrannosaur']
        rr_to_either = sorted(rr_to_comical + rr_to_dinosaur)
        self.assertEqual(sorted(related_rare_words(['comical', 'dinosaur'], sample_size=None)), rr_to_either)
        results = related_rare_words(['comical', 'dinosaur'])
        self.assertEqual(len(results), 8)
        self.assertTrue(set(rr_to_either).issuperset(set(results)))

    def test_related_rare_word(self):
        result_possibilities = ['artless', 'canticle', 'clownish', 'histrionic', 'humourous', 'parodic', 'risible',
                                'sophomoric', 'uneconomical', 'uproarious']
        self.assertIn(related_rare_word('comical'), result_possibilities)


class TestStochasticJolasticWordGenerator(unittest.TestCase):

    def test_random_nonrhyme(self):
        with open('tests/random_nonrhyme_possible_results.txt') as f:
            # There are over 5000 possible results even with rare words as input since this function sometimes calls
            # a random lexigen function on the result of a random lexigen function (and there are already ~70 possible
            # results even if only one function is called.
            possible_results = f.read().splitlines()
        markovgen = StochasticJolasticWordGenerator()
        for i in range(6):
            result = markovgen.random_nonrhyme(['pataphysics', 'Dadaist'])
            self.assertIn(result, possible_results)

    def test_nonlast_word(self):
        with open('tests/random_nonrhyme_possible_results.txt') as f:
            possible_randalg_results = f.read().splitlines()
        words_for_sampling = ['fervent', 'mutants', 'dazzling', 'flying', 'saucer', 'milquetoast']
        markovgen = StochasticJolasticWordGenerator()
        input_words = ['pataphysics', 'Dadaist']
        for i in range(2):
            result = markovgen.nonlast_word_of_markov_line(input_words[i:], words_for_sampling)
            self.assertTrue(result in possible_randalg_results or result in words_for_sampling or
                            result in markovgen.connector_choices)
            self.assertIn(markovgen.nonlast_word_of_markov_line(input_words[i:]), possible_randalg_results)

    def test_last_word(self):
        with open('tests/random_nonrhyme_possible_results.txt') as f:
            possible_randalg_results = f.read().splitlines()
        markovgen = StochasticJolasticWordGenerator()
        input_words = ['pataphysics', 'Dadaist']
        for i in range(2):
            result = markovgen.last_word_of_markov_line(input_words[i:], max_length=6)
            self.assertIn(result, possible_randalg_results)
            self.assertLessEqual(len(result), 6)
            rhyming_result = markovgen.last_word_of_markov_line(input_words[i:], rhyme_with='shudder', max_length=10)
            self.assertLessEqual(len(rhyming_result), 10)
            self.assertIn(rhyming_result, rhymes('shudder', sample_size=None))

class TestPoemGenerator(unittest.TestCase):

    def get_possible_word_list(self, input_word_list):
        possible_line_enders = ['.', ',', '!', '?', '...']
        possible_words = input_word_list.copy()
        for line_ender in possible_line_enders:
            for word in input_word_list:
                # Since we are testing using .split(), the list of possible words should include f'{word + line ender}'
                possible_words.append(word + line_ender)
        return possible_words

    def test_poem_line_from_word_list(self):
        input_word_list = ['crypt', 'crypts', 'crypt', 'ghost', 'ghosts', 'lost', 'time', 'times']
        possible_words = self.get_possible_word_list(input_word_list)
        possible_connectors = [',', '...', '&', 'and', 'or', '->']
        pgen = PoemGenerator()
        for i in range(5):  # Generate 5 random lines of poetry and test them.
            max_line_length = 35 + 5 * i
            poem_line = pgen.poem_line_from_word_list(input_word_list, max_line_length=max_line_length)
            # First character of line should not be a space as indents are handled by the poem_from_word_list function
            self.assertNotEqual(poem_line[0], ' ')
            # Should not have newlines as these are handled by the poem_from_word_list function
            self.assertNotIn('\n', poem_line)
            # When split, everything should derive from the possible word list
            self.assertTrue(set(possible_words + possible_connectors).issuperset(set(poem_line.split())))
            word, last_word = None, None
            for text in poem_line.split():
                word = re.match(r'[a-zA-Z]*', text).group()
                #  No word should be too similar to the preceding word
                if word and last_word:
                    self.assertFalse(too_similar(word, last_word))
                last_word = word
            # Line length should not exceed maximum line length
            self.assertTrue(len(poem_line) <= max_line_length)

    def test_poem_from_word_list(self):
        input_word_list = ['crypt', 'sleep', 'ghost', 'time']
        pgen = PoemGenerator()
        poems = [pgen.poem_from_word_list(input_word_list, limit_line_to_one_input_word=True),
                 pgen.poem_from_word_list(input_word_list, num_lines=8)]
        expected_newlines_in_poem = [5, 7]
        for i, poem in enumerate(poems):
            # 5 lines = 5 newline characters since one ends the poem
            self.assertEqual(poem.count('\n'), expected_newlines_in_poem[i])
            poem_lines = poem.split('\n')
            for string in poem_lines:
                indent_length = len(string) - len(string.lstrip())
                if indent_length != 0:
                    # Indent length should not repeat... unless there's no indent
                    self.assertNotEqual(indent_length, last_indent_length)
                last_indent_length = indent_length
            last_line_words = [word for word in poem_lines[-1].split(' ') if word != '']
            self.assertEqual(len(last_line_words), 2)
            self.assertIn(last_line_words[0], input_word_list[:-1])
            self.assertEqual(last_line_words[1], 'time')

    def test_poem_from_markov(self):
        input_words = ['chalice', 'crime', 'coins', 'spectacular', 'dazzle', 'enigma']
        pgen = PoemGenerator()
        poem = pgen.poem_from_markov(input_words=input_words, min_line_words=7, max_line_words=10, num_lines=8,
                                     max_line_length=66)
        self.assertEqual(len(poem.lines), 8)
        for line in poem.lines:
            words = line.split(" ")
            self.assertGreaterEqual(len(words), 7)
            self.assertLessEqual(len(words), 10)
            self.assertLessEqual(len(line), 71)


def test_poem_line_from_markov(self):
    pgen = PoemGenerator()
    pgen.currently_generating_poem = Poem(['pataphysics', 'Dadaist'], [])
    words_for_sampling = ['fervent', 'mutants', 'dazzling', 'flying', 'saucer', 'milquetoast']
    line = pgen.poem_line_from_markov('surrealist', num_words=8, rhyme_with=None,
                                      words_for_sampling=words_for_sampling, max_line_length=40)
    words = line.split(' ')
    self.assertLessEqual(len(line), 40)
    self.assertLessEqual(len(words), 8)
    markovgen = StochasticJolasticWordGenerator()
    self.assertNotIn(words[-1], markovgen.common_words)
    similarity_checks = list(itertools.combinations(words, 2))
    for word_pair in similarity_checks:
        self.assertFalse(too_similar(word_pair[0], word_pair[1]))
    line = pgen.poem_line_from_markov('surrealist', num_words=8, rhyme_with='bell',
                                      words_for_sampling=words_for_sampling, max_line_length=None)
    words = line.split(' ')
    self.assertEqual(len(words), 8)
    self.assertIn(words[-1], rhymes('bell', sample_size=None))
    self.assertNotIn(line.split(' ')[-1], markovgen.common_words)
    similarity_checks = list(itertools.combinations(words, 2))
    for word_pair in similarity_checks:
        self.assertFalse(too_similar(word_pair[0], word_pair[1]))
    line = pgen.poem_line_from_markov('surrealist', num_words=8, rhyme_with='unrhymable',
                                      words_for_sampling=words_for_sampling, max_line_length=None)
    words = line.split(' ')
    self.assertEqual(len(words), 8)
    self.assertNotIn(words[-1], markovgen.common_words)
    similarity_checks = list(itertools.combinations(words, 2))
    for word_pair in similarity_checks:
        self.assertFalse(too_similar(word_pair[0], word_pair[1]))


class TestPDFPNGGenerator(unittest.TestCase):

    def test_get_font_sizes(self):
        pdfgen = PDFGenerator()
        font_size = pdfgen.get_font_size('abra cadabra hocus pocus bananas eating locusts')  # 47 characters
        self.assertEqual(font_size, 16)
        font_size = pdfgen.get_font_size('this line is 26 characters')
        self.assertIn(font_size, [16, 18, 20])
        font_size = pdfgen.get_font_size('this line is short')
        self.assertIn(font_size, pdfgen.default_font_sizes)

    def test_get_max_x_coordinate(self):
        pdfgen = PDFGenerator()
        x_coordinate = pdfgen.get_max_x_coordinate('this line is 26 characters', 'Arial', 24)
        self.assertEqual(x_coordinate, 60)
        x_coordinate = pdfgen.get_max_x_coordinate('this line is thirty-five characters', 'Arial', 21)
        self.assertEqual(x_coordinate, 60)
        x_coordinate = pdfgen.get_max_x_coordinate('this line short', 'Courier', 18)
        self.assertEqual(x_coordinate, 60)
        x_coordinate = pdfgen.get_max_x_coordinate('this line 18 chars', 'Arial', 20)
        self.assertEqual(x_coordinate, 130)
        x_coordinate = pdfgen.get_max_x_coordinate('this line is 21 chars', 'Arial', 15)
        self.assertEqual(x_coordinate, 130)
        x_coordinate = pdfgen.get_max_x_coordinate('short', 'Arial', 15)
        self.assertEqual(x_coordinate, 280)
        pdfgen.orientation = 'portrait'
        x_coordinate = pdfgen.get_max_x_coordinate('this line 18 chars', 'Arial', 24)
        self.assertEqual(x_coordinate, 30)
        x_coordinate = pdfgen.get_max_x_coordinate('this line is thirty-five characters', 'Arial', 21)
        self.assertEqual(x_coordinate, 30)
        x_coordinate = pdfgen.get_max_x_coordinate('this line short', 'Courier', 18)
        self.assertEqual(x_coordinate, 30)
        x_coordinate = pdfgen.get_max_x_coordinate('this line 18 chars', 'Arial', 20)
        self.assertEqual(x_coordinate, 100)
        x_coordinate = pdfgen.get_max_x_coordinate('this line is 21 chars', 'Arial', 15)
        self.assertEqual(x_coordinate, 100)
        x_coordinate = pdfgen.get_max_x_coordinate('short', 'Arial', 15)
        self.assertEqual(x_coordinate, 250)

    def test_set_filename(self):
        pdfgen = PDFGenerator()
        input_words = ['chalice', 'crime', 'coins', 'spectacular', 'dazzle', 'enigma']
        self.assertEqual(pdfgen.set_filename(input_words), 'chalice,crime,coins,spectacular,dazzle,enigma.pdf')
        self.assertEqual(pdfgen.set_filename(input_words, file_extension='png'),
                         'chalice,crime,coins,spectacular,dazzle,enigma.png')

    # def test_generate_png(self):
    #     try:
    #         os.remove('tests/test.png')
    #     except Exception:
    #         pass
    #     pdfgen = PDFGenerator()
    #     pdfgen.generate_png('tests/test.pdf')
    #     self.assertEqual(os.path.isfile('tests/test.png'), True)
    #     os.remove('tests/test.png')


class TestChaoticConcretePoemPDFGenerator(unittest.TestCase):

    def test_generate_pdf(self):
        pdfgen = ChaoticConcretePoemPDFGenerator()
        test_input = 'chalice crime coins spectacular'
        with patch('builtins.input', return_value=test_input):
            pdfgen.generate_pdf()
        self.assertEqual(len(pdfgen.drawn_strings), 58)
        for ds in pdfgen.drawn_strings:
            self.assertGreaterEqual(ds.x, 15)
            self.assertLessEqual(ds.x, 440)
            self.assertGreaterEqual(ds.y, 15)
            self.assertLessEqual(ds.y, 800)
            self.assertIsNotNone(ds.font)
            self.assertIn(ds.font_size, pdfgen.default_font_sizes)
            self.assertTrue(type(ds.rgb), tuple)


class TestCharacterSoupPoemPDFGenerator(unittest.TestCase):

    def test_generate_pdf(self):
        pdfgen = CharacterSoupPoemPDFGenerator()
        pdfgen.generate_pdf()
        self.assertGreaterEqual(len(pdfgen.drawn_strings), 300)  # Random, varies
        for ds in pdfgen.drawn_strings:
            self.assertGreaterEqual(ds.x, 10)
            self.assertLessEqual(ds.x, 560)
            self.assertGreaterEqual(ds.y, 10)
            self.assertLessEqual(ds.y, 790)
            self.assertIsNotNone(ds.font)
            self.assertIn(ds.font_size, list(range(6, 73)))
            self.assertTrue(type(ds.rgb), tuple)


class TestStopwordSoupPoemPDFGenerator(unittest.TestCase):

    def test_generate_pdf(self):
        pdfgen = StopwordSoupPoemPDFGenerator()
        pdfgen.generate_pdf()
        self.assertGreaterEqual(len(pdfgen.drawn_strings), 157)
        for ds in pdfgen.drawn_strings:
            self.assertGreaterEqual(ds.x, 10)
            self.assertLessEqual(ds.x, 490)
            self.assertGreaterEqual(ds.y, 10)
            self.assertLessEqual(ds.y, 790)
            self.assertIsNotNone(ds.font)
            self.assertIn(ds.font_size, list(range(6, 41)))
            self.assertTrue(type(ds.rgb), tuple)


class TestMarkovPoemPDFGenerator(unittest.TestCase):

    def test_generate_pdf(self):
        pdfgen = MarkovPoemPDFGenerator()
        test_input = 'chalice crime coins spectacular'
        with patch('builtins.input', return_value=test_input):
            pdfgen.generate_pdf()
        y_coord = 550
        for ds in pdfgen.drawn_strings:
            self.assertGreaterEqual(ds.x, 15)
            self.assertLessEqual(ds.x, 250)
            self.assertEqual(ds.y, y_coord)
            self.assertIsNotNone(ds.font)
            self.assertIsNotNone(ds.font_size)
            self.assertTrue(type(ds.rgb), tuple)
            y_coord -= 32


class TestFuturistPoemPDFGenerator(unittest.TestCase):

    def test_generate_pdf(self):
        pdfgen = FuturistPoemPDFGenerator()
        test_input = 'chalice crime coins spectacular'
        with patch('builtins.input', return_value=test_input):
            pdfgen.generate_pdf()
        y_coord = 60
        for ds in pdfgen.drawn_strings:
            self.assertGreaterEqual(ds.x, 15)
            self.assertLessEqual(ds.x, 280)
            self.assertEqual(ds.y, y_coord)
            self.assertIsNotNone(ds.font)
            self.assertIsNotNone(ds.font_size)
            self.assertTrue(type(ds.rgb), tuple)
            y_coord += 31


import os
import random
import string
from os.path import isfile
from typing import List, TypeVar, Tuple
from reportlab.pdfgen import canvas
from generativepoetry.poemgen import *
from reportlab.pdfbase.pdfmetrics import registerFont, registerFontFamily
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.lib.pagesizes import letter, landscape
from nltk.corpus import stopwords
from pdf2image import convert_from_path, convert_from_bytes
from pdf2image.exceptions import (
    PDFInfoNotInstalledError,
    PDFPageCountError,
    PDFSyntaxError
)
from .utils import filter_word_list

rgb_tuple = Tuple[float]


class VisualPoemString():
    """The text drawn by reportlab at an XY coordinate--can be a line, a word, or just a character."""

    def __init__(self, text, x: int, y: int, font: str, font_size: int, rgb: Optional[Tuple] = None):
        self.text = text
        self.x = x
        self.y = y
        self.font = font
        self.font_size = font_size
        self.rgb = rgb


class PDFGenerator:
    default_font_sizes = [12, 14, 16, 18, 24, 32]
    font_choices = ['arial', 'arial-bold', 'arial-italic', 'arial-bolditalic',
                    'Courier-Bold', 'Courier-Bold', 'Courier-BoldOblique','Courier-BoldOblique',
                    'Helvetica', 'Helvetica-BoldOblique', 'Helvetica-Bold', 'Helvetica-Oblique',
                    'Times-Bold', 'Times-BoldItalic', 'Times-Italic', 'Times-Roman',
                    'Vera', 'VeraBd', 'VeraBI', 'VeraIt']

    def __init__(self):
        registerFont(TTFont('arial', 'arial.ttf'))
        registerFont(TTFont('arial-bold', 'arialbd.ttf'), )
        registerFont(TTFont('arial-italic', 'ariali.ttf'), )
        registerFont(TTFont('arial-bolditalic', 'arialbi.ttf'), )
        registerFont(TTFont('Vera', 'Vera.ttf'))
        registerFont(TTFont('VeraBd', 'VeraBd.ttf'))
        registerFont(TTFont('VeraIt', 'VeraIt.ttf'))
        registerFont(TTFont('VeraBI', 'VeraBI.ttf'))
        self.orientation = 'landscape'
        self.drawn_strings: List[VisualPoemString] = []

    def get_font_size(self, line):
        if len(line) > 30:
            return 16
        elif len(line) >= 24:
            return random.choice([16, 18, 20])
        else:
            return random.choice(self.default_font_sizes)

    def get_max_x_coordinate(self, line, font_choice, font_size):
        if (font_size >= 23 and len(line) >= 17) or (font_size >= 20 and len(line) > 30) or \
                font_choice.startswith('Courier'):  # Courier is the widest
            return 30 if self.orientation == 'portrait' else 60
        elif (font_size == 23 and len(line) > 14) or (font_size >= 20 and len(line) > 16) or len(line) > 20:
            return 100 if self.orientation == 'portrait' else 130
        else:
            return 250 if self.orientation == 'portrait' else 280

    def set_filename(self, input_words, file_extension='pdf'):
        sequence = ""
        filename = f"{','.join(input_words)}{sequence}.{file_extension}"
        while isfile(filename):
            sequence = int(sequence or 0) + 1
            filename = f"{','.join(input_words)}({sequence}).{file_extension}"
        self.pdf_filepath = os.getcwd() + '/'  + filename
        return filename

    def generate_png(self, input_filepath=None):
        print('Generating image...')
        pages = convert_from_path(input_filepath)
        for page in pages:
            page.save(f'{input_filepath[:-3]}png', 'PNG')


class ChaoticConcretePoemPDFGenerator(PDFGenerator):

    def generate_pdf(self, input_words: Optional[List[str]] = [], max_words=Optional[int]):
        self.drawn_strings = []
        input_words = get_input_words() if not len(input_words) else input_words
        output_words = input_words + phonetically_related_words(input_words)
        random.shuffle(output_words)
        filename = self.set_filename(input_words)
        c = canvas.Canvas(filename)
        for word in output_words[:200]:
            word = random.choice([word, word, word, word.upper()])
            x = random.randint(15,440)
            y = random.randint(15, 800)
            font_choice = random.choice(self.font_choices)
            font_size = random.choice(self.default_font_sizes)
            rgb = get_random_color()
            vp_string = VisualPoemString(word, x=x, y=y, font=font_choice, font_size=font_size, rgb=rgb)
            c.setFont(vp_string.font, vp_string.font_size)
            c.setFillColorRGB(*vp_string.rgb)
            c.drawString(vp_string.x, vp_string.y, vp_string.text)
            self.drawn_strings.append(vp_string)
        c.showPage()
        c.save()


class CharacterSoupPoemPDFGenerator(PDFGenerator):

    def generate_pdf(self):
        c = canvas.Canvas("character_soup.pdf")
        for i in range(20):
            char_sequence = random.choice([string.ascii_lowercase, string.digits, string.punctuation])
            for char in char_sequence:
                char = random.choice([char, char, char.upper(), char.upper()]) \
                    if char_sequence == string.ascii_lowercase else char
                font_choice = random.choice(self.font_choices)
                font_size = random.randint(6, 72)
                x = random.randint(10, 560)
                y = random.randint(10, 790)
                rgb = get_random_color()
                vp_string = VisualPoemString(char, x=x, y=y, font=font_choice, font_size=font_size, rgb=rgb)
                c.setFillColorRGB(*rgb)
                c.setFont(vp_string.font, vp_string.font_size)
                c.drawString(vp_string.x, vp_string.y, vp_string.text)
                self.drawn_strings.append(vp_string)
        c.showPage()
        c.save()


class StopwordSoupPoemPDFGenerator(PDFGenerator):
    default_font_sizes = [6, 16, 24, 32, 48]

    def generate_pdf(self):
        c = canvas.Canvas("stopword_soup.pdf")
        punctuation = [char for char in string.punctuation]
        words_to_use = filter_word_list(stopwords.words('english')) + punctuation  # filter removes 1/2s of contractions
        for word in ['him', 'her', 'his', 'they', 'won']:
            words_to_use.remove(word)
        for i in range(3):
            words_to_use.extend(['hmm', 'ah', 'umm', 'uh', 'ehh.', 'psst..', 'what?', 'oh?', 'ahem..'])
        random.shuffle(words_to_use)
        for word in words_to_use:
            word = random.choice([word, word, word.upper(), word.upper()])
            font_choice = random.choice(self.font_choices)
            font_size = random.randint(6, 40)
            x = random.randint(10, 490)
            y = random.randint(10, 790)
            rgb = get_random_color(threshold=.5)
            c.setFillColorRGB(*rgb)
            vp_string = VisualPoemString(word, x=x, y=y, font=font_choice, font_size=font_size, rgb=rgb)
            c.setFont(vp_string.font, vp_string.font_size)
            c.drawString(vp_string.x, vp_string.y, vp_string.text)
            self.drawn_strings.append(vp_string)
        c.showPage()
        c.save()


class MarkovPoemPDFGenerator(PDFGenerator):
    default_font_sizes = [15, 18, 21, 24, 28]

    def generate_pdf(self, input_words: Optional[List[str]] = [], orientation: string = 'landscape'):
        self.drawn_strings = []
        self.orientation = orientation
        if self.orientation.lower() == 'landscape':
            num_lines = 14
            y_coordinate = 550
            min_line_words = 7
            max_line_words = 10
            max_line_length = 66
            min_x_coordinate = 60
        elif self.orientation.lower() == 'portrait':
            num_lines = 24
            y_coordinate = 740
            min_line_words = 4
            max_line_words = 7
            max_line_length = 40
            min_x_coordinate = 15
        else:
            raise Exception('Must choose from the following orientations: portrait, landscape')
        input_words = get_input_words() if not len(input_words) else input_words
        poemgen = PoemGenerator()
        poem = poemgen.poem_from_markov(input_words=input_words, min_line_words=min_line_words, num_lines=num_lines,
                                        max_line_words=max_line_words, max_line_length=max_line_length)
        font_choice, last_font_choice = None, None
        filename = self.set_filename(input_words)
        if orientation == 'landscape':
            c = canvas.Canvas(filename, pagesize=landscape(letter))
        else:
            c = canvas.Canvas(filename)
        for line in poem.lines:
            text = random.choice([line, line, line, line.upper()])
            font_size = self.get_font_size(line)
            while font_choice is None or last_font_choice == font_choice:
                font_choice = random.choice(self.font_choices)
            max_x_coordinate = self.get_max_x_coordinate(line, font_choice, font_size)
            last_font_choice = font_choice
            x_coordinate = random.randint(min_x_coordinate, max_x_coordinate)
            vp_string = VisualPoemString(line, x=x_coordinate, y=y_coordinate, font=font_choice, font_size=font_size)
            c.setFont(vp_string.font, vp_string.font_size)
            c.drawString(vp_string.x, vp_string.y, vp_string.text)
            y_coordinate -= 32
            self.drawn_strings.append(vp_string)
        c.showPage()
        c.save()


class FuturistPoemPDFGenerator(PDFGenerator):
    connectors = [' + ', ' - ', ' * ', ' % ', ' = ', ' != ', ' :: ']
    default_font_sizes = [15, 18, 21, 24, 28]

    def generate_pdf(self, input_words: Optional[List[str]] = [],):
        self.drawn_strings = []
        input_words = get_input_words() if not len(input_words) else input_words
        word_list = input_words + phonetically_related_words(input_words)
        poem_lines = []
        pgen = PoemGenerator()
        for i in range(25):
            random.shuffle(word_list)
            poem_lines.append(pgen.poem_line_from_word_list(word_list, connectors=self.connectors, max_line_length=40))
        filename = self.set_filename(input_words)
        c = canvas.Canvas(filename)
        y_coordinate = 60
        for line in poem_lines:
            line = random.choice([line, line, line, line.upper()])
            font_choice = random.choice(self.font_choices)
            font_size = self.get_font_size(line)
            max_x_coordinate = self.get_max_x_coordinate(line, font_choice, font_size)
            x_coordinate = random.randint(15, max_x_coordinate)
            vp_string = VisualPoemString(line, x=x_coordinate, y=y_coordinate, font=font_choice, font_size=font_size)
            c.setFont(vp_string.font, vp_string.font_size)
            c.drawString(vp_string.x, vp_string.y, line)
            y_coordinate += 31
            self.drawn_strings.append(vp_string)
        c.showPage()
        c.save()